#ifndef _SERVICE_H_
#define _SERVICE_H_

#include "../../pfc/pfc.h"

#include "utf8api.h"

#include "core_api.h"

long interlocked_increment(long * var);//note: win32 sucks, return values arent reliable, they can be used only to determine if new value is <0, 0 or >0
long interlocked_decrement(long * var);

class service_owner;

class NOVTABLE service_base		//ALL service classes MUST inherit from this
{
private:
	service_owner * owner;
	long refcount;
protected:
	service_base() : owner(0), refcount(1) {}
	~service_base() {}
public:	
	virtual int service_release();
	virtual int service_add_ref();
	virtual service_base * service_query(const GUID & guid) {return 0;}//if you override, always call parent class for unknown guids
	inline void service_set_owner(service_owner * p_owner) {assert(owner==0);owner=p_owner;}//used only internally !
	//void service_release_safe();
	inline void service_release_safe() {service_release();}
};

class NOVTABLE service_base_static : public service_base
{
public:
	virtual int service_release() {return 1;}
	virtual int service_add_ref() {return 1;}
};

/*
//msvc6 sucks

template<class T>
	static T* service_query_t(service_base * ptr) {return static_cast<T*>(ptr->service_query(T::get_class_guid()));}

*/
#define service_query_t(T,ptr) (static_cast<T*>(ptr->service_query(T::get_class_guid())))
//user service_release on returned object when youre done

class NOVTABLE service_owner
{
public:
	virtual service_base * instance_create()=0;
	virtual void instance_release(service_base * ptr)=0;
};

//some things use service instances that arent created through registered factories
template<class T>
class service_owner_t : public service_owner
{
public:
	virtual service_base * instance_create() {service_base * ptr = dynamic_cast<service_base*>(new T); ptr->service_set_owner(this); return ptr; }
	virtual void instance_release(service_base * ptr) {delete static_cast<T*>(ptr);}
	T* create() {return static_cast<T*>(instance_create());}
};

template<class T>
class service_owner_single_t : public service_owner
{
	T instance;
public:
	service_owner_single_t() {instance.set_owner(this);}
	virtual service_base * instance_create() {service_base * ptr = dynamic_cast<service_base*>(&instance); ptr->service_add_ref(); return ptr; }
	virtual void instance_release(service_base * ptr) {assert(static_cast<T*>(ptr)==&instance);}
	T* create() {return static_cast<T*>(instance_create());}
};

class NOVTABLE service_factory_base : public service_owner
{
private:
	static service_factory_base *list;
	service_factory_base * next;	
	GUID guid;
protected:
	service_factory_base(const GUID & p_guid) {assert(!core_api::are_services_available());guid=p_guid;next=list;list=this;}
public:

	//moved to service_owner
	//virtual service_base * instance_create()=0;
	//virtual void instance_release(service_base * ptr)=0;
	
	inline const GUID & get_class_guid() {return guid;}

	inline static service_factory_base * list_get_pointer() {return list;}
	inline service_factory_base * list_get_next() {return next;}

	static service_base * enum_create(const GUID &g,int n);
	static int enum_get_count(const GUID &g);
	inline static bool is_service_present(const GUID & g) {return enum_get_count(g)>0;}

#ifdef FOOBAR2000_EXE
	static void process_components_directory(const char * path,service_factory_base ** & blah);
	static void on_app_init(const char * exe_path);
	static void on_app_shutdown();
	static void config_reset(const char * name = 0);
	static void on_app_post_init();
	static void on_saveconfig(bool b_reset=false);
#endif

	//obsolete, use core_api
	inline static HINSTANCE get_my_instance() {return core_api::get_my_instance();}
	inline static const char * get_my_file_name() {return core_api::get_my_file_name();}
	inline static const char * get_my_full_path() {return core_api::get_my_full_path();}
	inline static HWND get_main_window() {return core_api::get_main_window();}

};


/*

//msvc6 sucks

template<class T>
	static T * service_enum_create_t(int n) {return (T*)service_factory_base::enum_create(T::get_class_guid(),n);}
template<class T>
	static int service_enum_get_count_t() {return service_factory_base::enum_get_count(T::get_class_guid());}
*/

#define service_enum_create_t(T,n) (static_cast<T*>(service_factory_base::enum_create(T::get_class_guid(),n)))
#define service_enum_get_count_t(T) (service_factory_base::enum_get_count(T::get_class_guid()))
#define service_enum_is_present(g) (service_factory_base::is_service_present(g))
#define service_enum_is_present_t(T) (service_factory_base::is_service_present(T::get_class_guid()))


class service_enum
{
private:
	int service_ptr;
	GUID guid;
public:
	service_enum(const GUID &);
	void reset();
	service_base * first() {reset();return next();}
	service_base * next();
};

template<class B>
class service_enum_t : private service_enum
{
public:
	service_enum_t() : service_enum(B::get_class_guid()) {}
	void reset() {service_enum::reset();}
	B * first() {return (B*)(service_enum::first());}
	B * next() {return (B*)(service_enum::next());}
};


template<class B,class T>
class service_factory_t : public service_factory_base
{
#ifdef _DEBUG
	long issued;
#endif
public:
	service_factory_t() : service_factory_base(B::get_class_guid())
#ifdef _DEBUG
		, issued(0)
#endif
	{}
#ifdef _DEBUG
	~service_factory_t() {assert(!issued);}
#endif

	virtual service_base * instance_create()
	{
#ifdef _DEBUG
		interlocked_increment(&issued);
#endif
		service_base * ptr = dynamic_cast<service_base*>(dynamic_cast<B*>(new T));
		ptr->service_set_owner(this);
		return ptr;
	}
	virtual void instance_release(service_base * ptr)
	{
#ifdef _DEBUG
		interlocked_decrement(&issued);
#endif
		delete static_cast<T*>(static_cast<B*>(ptr));
	}
};

template<class B,class T>
class service_factory_single_t : public service_factory_base
{
private:
	T instance;
public:
	service_factory_single_t() : service_factory_base(B::get_class_guid()) {instance.service_set_owner(this);}

	virtual service_base * instance_create()
	{
		service_base * ret = dynamic_cast<service_base*>(dynamic_cast<B*>(&instance));
		ret->service_add_ref();
		return ret;
	}

	T& get_static_instance() {return instance;}

	virtual void instance_release(service_base * ptr) {assert(0);}
};

template<class T>
class service_factory_single_ref_t : public service_factory_base
{
private:
	T & instance;
public:
	service_factory_single_ref_t(T& param) : instance(param), service_factory_base(T::get_class_guid()) {instance.service_set_owner(this);}

	virtual service_base * instance_create()
	{
		service_base * ret = dynamic_cast<service_base*>(&instance);
		ret->service_add_ref();
		return ret;
	}

	T& get_static_instance() {return instance;}

	virtual void instance_release(service_base * ptr) {assert(0);}
};

class service_single_base : public service_base, private service_factory_base
{
	virtual service_base * instance_create()
	{
		service_base * ptr = dynamic_cast<service_base*>(this);
		ptr->service_add_ref();
		return ptr;
	}
	virtual void instance_release(service_base * ptr) {assert(0);}
protected:
	service_single_base(const GUID & param) : service_factory_base(param) {}
};


#endif